"""
Author:      www.tropofy.com

Copyright 2013 Tropofy Pty Ltd, all rights reserved.

This source file is part of Tropofy and govered by the Tropofy terms of service
available at: http://www.tropofy.com/terms_of_service.html

This source file is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the license files for details.
"""

from datetime import datetime, date, time
from sqlalchemy.exc import IntegrityError
from tropofy.database.tropofy_orm import ValidationException, Validator
from tropofy.widgets.widgets import Widget, WidgetConfigException


class Form(Widget):
    """A form through which data can be added and edited."""

    class FormElementException(Exception):
        pass

    class Element(object):
        """An element of a form.

        :param name: Form element name. Used to reference the value of the element on submit.
        :type name: str
        :param input_type: Type of form input. (Form.Element.TEXT, Form.Element.SELECT, Form.Element.DATE, Form.Element.TIME, Form.Element.DATETIME, Form.Element.BOOL)
        :type input_type: str
        :param default: (optional) Default input value. Default defaults are: TEXT:"",  SELECT: value of first option. If SELECT element, default must be an option value.
        :type default: Multiple possible types. Must be valid relative to options list and input_type.
        :param label: (optional) Label of form element.
        :type label: str
        :param hidden: (optional) If true, form element will be hidden
        :type hidden: bool
        :param disabled: (optional) If true, form element will be disabled
        :type disabled: bool
        :param options: (optional - used if input_type=='select', or 'typeahead') for select valid options that this element may take, for typeahead suggested options that this element may take
        :type options: for select expects list of strings or list of :class:`tropofy.widgets.Form.Element.Option`, for typeahead expects a list of strings
        :param compute_options_from: (optional - used if input_type=='select' and options=None). String matching the `name` of a dependent_option_list of a :class:`tropofy.Form.Element.Option` within a :class:`tropofy.Form.Element` in this widget.
        :type compute_options_from: str
        """
        TEXT = 'text'
        SELECT = 'select'
        DATE = 'date'
        TIME = 'time'
        DATETIME = 'datetime'
        BOOL = 'bool'
        TYPEAHEAD = 'typeahead'

        VALID_INPUT_TYPES = [TEXT, SELECT, DATE, TIME, DATETIME, BOOL, TYPEAHEAD]
        VALID_DATETIME_INPUT_TYPES = {DATE, TIME, DATETIME}

        def __init__(self, name, input_type, default=None, label=None, hidden=False, disabled=False, options=None, compute_options_from=None):
            if input_type == Form.Element.BOOL:
                input_type = Form.Element.SELECT
                options = [
                    Form.Element.Option(value=True, text='True'),
                    Form.Element.Option(value=False, text='False'),
                ]
                default = default if default is not None else True
                if not isinstance(default, bool):
                    raise Form.FormElementException("Invalid default for boolean: '%s'. Must be either True or False" % default)

            self.name = name
            self.input_type = input_type  # Must be defined by options. Options validator uses this.
            if self.is_select_element:
                if compute_options_from is not None:
                    # self.options = [] # Must be defined before default_value. Default validator uses this.
                    self._options = []
                    self.compute_options_from = compute_options_from
                else:
                    self.options = options
                    self.compute_options_from = None
            else:
                self.options = options
                self.compute_options_from = None

            self.default_value = default
            self.hidden = hidden
            self.disabled = disabled
            self.label = label if label is not None else self.name
            # self.compute_options_from = compute_options_from if self.options == [] and self.is_select_element else None

        def _serialise(self):
            return {
                'name': self.name,
                'inputType': self.input_type,
                'defaultValue': self.default_value,
                'options': [o._serialise() for o in self.options] if self.is_select_element else self.options,
                'computeOptionsFrom': self.compute_options_from,
                'label': self.label,
                'hidden': self.hidden,
                'disabled': self.disabled,
            }

        @property
        def is_select_element(self):
            return self.input_type == Form.Element.SELECT

        @property
        def is_typeahead_element(self):
            return self.input_type == Form.Element.TYPEAHEAD

        ####### INPUT VALIDATORS #############################
        @property
        def input_type(self):
            return self._input_type

        @input_type.setter
        def input_type(self, value):
            if value not in Form.Element.VALID_INPUT_TYPES:
                raise Form.FormElementException("Invalid input type '%s'. Must be one of %s" % (value, str(Form.Element.VALID_INPUT_TYPES)))
            self._input_type = value

        @property
        def options(self):
            return self._options

        @options.setter
        def options(self, value):
            self._options = []
            if self.is_select_element:
                for option in value:
                    if isinstance(option, Form.Element.Option):
                        self._options.append(option)
                    else:
                        if isinstance(option, basestring):
                            option = {'value': option}
                        try:
                            self._options.append(Form.Element.Option(**option))
                        except:
                            raise Form.FormElementException('Could not create Form.Element.Option from %s. You should create Form.Element.Option mixins directly in your app. Refer to docs for details.' % option)
                if len(self._options) < 1:
                    raise Form.FormElementException("Select element '%s' has no options. Please pass a list of options." % self.name)

            elif self.is_typeahead_element:
                self._options = value

        @property
        def default_value(self):
            return self._default_value

        @default_value.setter
        def default_value(self, value):
            if value is None:
                if self.input_type in [Form.Element.TEXT, Form.Element.TYPEAHEAD]:
                    value = ""

                elif self.input_type in [Form.Element.DATE, Form.Element.TIME, Form.Element.DATETIME]:
                    value = ""
                elif self.is_select_element and len(self.options) > 0:
                    value = self.options[0].value  # Checked earlier that if SELECT element, at least one option.
                elif self.compute_options_from is not None:
                    value = ""
                else:
                    raise Form.FormElementException("Default value not provided, and no 'default default' for form element type %s" % self.input_type)

            if self.is_select_element:
                option_values = [o.value for o in self.options]
                if value not in option_values and self.compute_options_from is not None:
                    if len(option_values) > 0:
                        self._default_value = option_values[0]
                elif value not in option_values:
                    raise Form.FormElementException("Default for `select` element exception. Default value '%s' not found in options values: %s" % (value, str(option_values)))
                else:
                    pass

            self._default_value = value
        ######## END INPUT VALIDATORS #########################


        def _process_options(self, options):
            """Convert to list of :class:`tropofy.widgets.Form.Element.Option` from option formats:

            Options allowed to be input as list of strings or list of dictionaries - {value: <value>, text: <text>},
            where <text> is the text displayed in the select and <value> the value submitted with the form.

            :rtype: list of :class:`tropofy.widgets.Form.Element.Option`
            """



        class Option(object):
            """An option of a form select.

            :param value: The value of this option submitted with the form if selected.
            :type value: jsonifiable type (usually str or int)
            :param text: (Optional) Option text displayed in select list. Default to value if not provided.
            :type text: str
            :param dependent_option_lists:
            :type dependent_option_lists: list of dicts of form {name: <option list name> options: <list of :class:`tropofy.widgets.Form.Element.Option`>
            """
            def __init__(self, value, text=None, dependent_option_lists=None):
                self.value = value
                self.text = text if text else self.value
                self.dependent_option_lists = dependent_option_lists if dependent_option_lists else {}

            def _serialise(self):
                return {
                    'value': self.value,
                    'text': self.text,
                    'dependentOptionLists': [{
                        'name': name,
                        'options': [o._serialise() for o in options]
                    } for name, options in self.dependent_option_lists.iteritems()]
                }

    def _get_type(self):
        return "FormWidget"

    def _refresh(self, request, data_set, **kwargs):  # Refresh the entire form.
        return self.refresh(request, data_set, **kwargs)

    def _update(self, request, data, oper, data_set, **kwargs):  # Submit form
        if oper == 'submitFormElements':
            form_elements = self.get_form_elements(data_set)
            results = []
            for element in [e for e in form_elements if data.get(e.name)]:
                try:
                    if element.input_type == Form.Element.DATE:
                        data[element.name] = Validator.validate(data[element.name], date, element.name)
                    elif element.input_type == Form.Element.TIME:
                        data[element.name] = Validator.validate(data[element.name], time, element.name)
                    elif element.input_type == Form.Element.DATETIME:
                        data[element.name] = Validator.validate(data[element.name], datetime, element.name)
                except (ValidationException, IntegrityError) as e:
                    results.append({
                        'name': element.name,
                        'success': False,
                        'message': e.message if e.message else 'A validation error occurred.'
                    })
            if not results:  # No exception
                results = self.process_data(data, data_set)
            return results if results else ""
        elif oper == 'updateForm':
            form_values = data
            form_elements = self.get_form_elements(data_set, form_values=form_values)
            form_elements = form_elements if form_elements else []

            return {
                'formElements': [e._serialise() for e in form_elements],
                # 'formTitle': self.get_title(data_set, form_values=form_values),  // TODO: Need to add **kwargs before can pass form_values.
                # 'isAutoSubmit': self.is_autosubmit(data_set, form_values=form_values),
                # 'isSubmitEntireForm': self.is_submit_entire_form(data_set, form_values=form_values),
                # 'hideSubmitButton': self.hide_submit_button(data_set, form_values=form_values),
            }

    def refresh(self, request, data_set, **kwargs):
        form_elements = self.get_form_elements(data_set)
        form_elements = form_elements if form_elements else []

        return {
            'formElements': [e._serialise() for e in form_elements],
            'formTitle': self.get_title(data_set),
            'isAutoSubmit': self.is_autosubmit(data_set),
            'isSubmitEntireForm': self.is_submit_entire_form(data_set),
            'hideSubmitButton': self.hide_submit_button(data_set),
            'datetimeDisplayFormats': self._get_datetime_display_formats(),
        }

    def get_form_elements(self, data_set, **kwargs):
        """Return a list of elements to use in the form.

        :kwarg form_values: If provided, is a dict of form element {name: value}. Only passed if this function is being called when the form exists on the webpage, not for an initial load or refresh.

        :rtype: list of :class:`tropofy.widgets.Form.Element`
        """
        raise NotImplementedError

    def process_data(self, data, data_set, **kwargs):
        """Process data from form submit.

        :param data: Dict of name:value pairs, submitted from the form.
        :type data: dict
        :param data_set: current data set.
        :type data_set: :class:`tropofy.app.AppDataSet`
        :returns: Dictionary in the following form - {'success': <bool>, 'message': <HTML string>, 'results': <results_list>}

            - success (optional bool): Indicates full form submit success or failure.
            - message (optional HTML string): Response message for full form submit.
            - results (optional list of dicts of the form): {'name': name, 'success': <bool>, 'message': <str>}.
               - name (str): Must match the name of a corresponding :class:`tropofy.widgets.Form.Element`.
               - success (optional bool): defaults to true. Indicates single form element submit success or failure.
               - message (optional str): default to null. Response message for single form element submit.

        'name', 'success', 'message', 'value'
        :rtype: dict
        """
        raise NotImplementedError

    def get_title(self, data_set, **kwargs):
        """A title for the form. If not implemented then there will be no title.

        :param data_set: current data set.
        :type data_set: :class:`tropofy.app.AppDataSet`
        :returns: Title for the form.
        :rtype: str
        """
        return None

    def is_autosubmit(self, data_set, **kwargs):
        """Determines whether a form autosubmits on value change or has a submit button.

        - True (default): The entire form will autosubmit when any value is changed. No submit button will be shown.
        - False: A submit button will be shown and the form will submit only when the button is clicked.

        :param data_set: current data set.
        :type data_set: :class:`tropofy.app.AppDataSet`
        :returns: Boolean determining if the form will autosubmit.
        :rtype: bool
        """
        return True

    def is_submit_entire_form(self, data_set, **kwargs):
        """Determines whether a form on submit submits the entire form.

        Default False - means only form element that has changed is submitted.
        :param data_set: current data set.
        :type data_set: :class:`tropofy.app.AppDataSet`
        :returns: Boolean determining if autosubmit form with submit entire form.
        :rtype: bool
        """
        return False

    def hide_submit_button(self, data_set, **kwargs):
        """Hides the form submit button. Advanced use only - it is usually safe to rely on default value.

        By default, submit button is hidden if :func:`tropofy.widgets.Form.is_autosubmit` returns True, and False otherwise.

        :param data_set: current data set.
        :type data_set: :class:`tropofy.app.AppDataSet`
        :returns: Boolean determining if the form submit button is hidden.
        :rtype: bool
        """
        return self.is_autosubmit(data_set)

    def _get_datetime_display_formats(self):
        formats = {
            Form.Element.DATE: "DD/MM/YYYY",
            Form.Element.DATETIME: "HH:mm DD/MM/YYYY",
            Form.Element.TIME: "HH:mm"
        }
        custom_formats = self.get_datetime_display_formats()
        if set(custom_formats.keys()) > Form.Element.VALID_DATETIME_INPUT_TYPES:
            raise WidgetConfigException("Error in function Form.get_datetime_display_formats. Must return dict with keys in: [Form.Element.DATE, Form.Element.TIME, Form.Element.DATETIME]")
        formats.update(custom_formats)
        return formats

    def get_datetime_display_formats(self, **kwargs):
        """ Specify custom display formats to display datetimes, dates, and times in a form.

        Use format strings specified by `MomentJS <http://momentjs.com/docs/#/parsing/string-format/>`_. Note that these are slightly different to string-formats used in Python.

        Allowable keys are: Form.Element.DATE, Form.Element.DATETIME, and Form.Element.DATE

        :returns: dict
        """
        return {}

    @classmethod
    def _get_input_type_from_python_type(cls, input_type):
        """Helper function to match python types with form input type.

        Useful in :func:`tropofy.widgets.Form.get_form_elements`, when creating :class:`tropofy.widgets.Form.Element`

        :param input_type: A Python type (eg. int, str, float, bool, date, time)
        :type input_type: type
        :returns: A form input type. One of 'text', 'select', 'data', 'time'.
        :rtype: str
        """
        type_to_form_type = {
            int: cls.Element.TEXT,
            str: cls.Element.TEXT,
            float: cls.Element.TEXT,
            bool: cls.Element.BOOL,
            date: cls.Element.DATE,
            time: cls.Element.TIME,
            datetime: cls.Element.DATETIME,
        }
        return type_to_form_type.get(input_type, Form.Element.TEXT)


class ParameterForm(Form):
    """An implementation of :class:`tropofy.widgets.Form` that displays all :class:`tropofy.app.Parameter` defined in :func:`tropofy.app.AppWithDataSets.get_parameters`

    :param parameter_names_filter: (Optional) List of parameter names to display on form. If none are specified, default is to display all parameters.
    :type parameter_names_filter: list
    """
    def __init__(self, title='Parameters', parameter_names_filter=None):
        self.parameter_names_filter = parameter_names_filter if parameter_names_filter is not None else []
        self.title = title

    def get_form_elements(self, data_set, **kwargs):
        elements = []
        for param in data_set.app.get_parameters():
            if not self.parameter_names_filter or param.name in self.parameter_names_filter:
                value = data_set.get_param(param.name)
                options = param.options

                elements.append(Form.Element(
                    name=param.name,
                    input_type=Form.Element.SELECT if options else ParameterForm._get_input_type_from_python_type(param.allowed_type if param.allowed_type else type(value)),
                    default=value,
                    label=param.label,
                    options=options,
                ))
        return elements

    def get_title(self, data_set, **kwargs):
        return self.title

    def process_data(self, data, data_set, **kwargs):
        results = []
        for name, value in data.iteritems():
            try:
                data_set.set_param(name=name, value=value)
            except (ValidationException, IntegrityError) as e:
                results.append({
                    'name': name,
                    'success': False,
                    'message': e.message if e.message else 'A validation error occurred.'
                })
        return {'results': results}


class Filter(Form):
    """An implementation of :class:`tropofy.widgets.Form` specifically designed to act as a filter for other widgets.

    Only the method :func:`tropofy.widgets.Filter.get_form_elements` needs to be implemented for this widget.

    Pass as the first parameter to a :class:`tropofy.widgets.FilteredChart` to link the filter.
    """

    def get_form_elements(self, data_set, **kwargs):
        """Return a list of :class:`tropofy.widgets.Form.Element`

        When implementing :func:`tropofy.widgets.Form.get_form_elements`, ensure the :func:`tropofy.widgets.Form.Element` default values are that which you want your
        filtered widget to use on first use.

        :param data_set: The DataSet object on which queries can be made to access the apps data.
        :type data_set: :class:`tropofy.app.AppDataSet`
        :rtype: list of :class:`tropofy.widgets.Form.Element`
        """
        raise NotImplementedError

    @property
    def var_name(self):
        return 'filter.%s' % self.id

    def get_values(self, data_set):
        """Values of filter, indexed by form element name.

        It is possible to overide this method and return custom, computed values. For example, if the form has separate input date, and time fields, these could be combined for a datetime value.
        If doing this, it is requiered that all form values be returned also, and not replaced. This is necessary to ensure the form updates on reload appropriately.

        :returns: dict
        """
        values = {}
        for element in self.get_form_elements(data_set):  # Would be a way to make this faster by storing just the names somewhere other than get_form_elements. Would probably need an extra use fn though.
            value = data_set.get_var(self._get_var_name(element.name))  # Don't return a None value, just don't return the key!
            if value is not None:
                values.update({
                    element.name: value
                })
        return values

    def process_data(self, data, data_set, **kwargs):
        for name, value in data.items():
            data_set.set_var(self._get_var_name(name), value)

    def _get_var_name(self, element_name):
        return 'filter_%s_%s' % (self.id, element_name)

    def is_submit_entire_form(self, data_set, **kwargs):
        return True

    def _refresh(self, request, data_set, **kwargs):
        if not self.get_values(data_set):
            self._set_filter_var_to_defaults(data_set)

        form_element_values = self.get_values(data_set)
        form_elements = self.get_form_elements(data_set)
        for element in form_elements:
            element.default_value = form_element_values[element.name]

        refresh_dict = super(Filter, self).refresh(request, data_set)
        refresh_dict['formElements'] = [e._serialise() for e in form_elements]  # Have to overide here, as only other option would be to pass form_elements to Form.refresh -> This is a framework function though, so this would be weird...
        return refresh_dict

    def _set_filter_var_to_defaults(self, data_set):
        # Need to set var to defaults on first load. Otherwise filtered widget will try and load and var will be empty.
        defaults = dict((fe.name, fe.default_value) for fe in self.get_form_elements(data_set))
        self.process_data(
            data=defaults,
            data_set=data_set
        )


class FilterMixin(object):
    """FilterMixin can be inherited from to link a widget to a :class:`tropofy.widgets.Filter`.

    Any widget inheriting from FilterMixin gets a constructor with a ``data_filter`` parameter and the helper function :func:`tropofy.widgets.FilteredChart.get_filter_values`.

    If you want to provide your own __init__, be sure to call :func:`tropofy.widgets.FilteredWidget.__init__` from within it.

    .. note:: It is important to inherit from FilterMixin before inheriting from another Widget. Otherwise the relationship to the filter may not work correctly.

    :param data_filter: A filter widget that filters this widget.
    :type data_filter: :class:`tropofy.widgets.Filter`
    """
    def __init__(self, data_filter):
        self.filter = data_filter
        if self.filter:
            self.filter.add_event_subscription('refreshed', self.actions('show'))  # After filter initially loaded('refreshed'), show the chart.
            self.filter.add_event_subscription('formSubmitSuccess', self.actions('refresh'))

    def get_filter_values(self, data_set, **kwargs):
        """Helper method that returns the values of the filter.

        Use this method within any methods of your widget with ``self.get_filter_values(data_set)``.

        :param data_set: The DataSet object on which queries can be made to access the apps data.
        :type data_set: :class:`tropofy.app.AppDataSet`

        :rtype: dict representation of the filtered values
        """
        try:
            return self.filter.get_values(data_set)
        except AttributeError:  # No filter.
            return {}

    def is_hidden(self, **kwargs):  # Makes widget hidden onload. Filter will trigger to show it when ready. Stops chart getting loaded incorrectly before filtered vars are set. If no filter, no need to hide chart.
        return self.filter is not None